@@ -20,6 +20,10 @@ extern UV  nth_prime_approx(UV x);
 extern UV  prime_count_lower(UV x);
 extern UV  prime_count_upper(UV x);
 extern UV  prime_count_approx(UV x);
+extern UV  twin_prime_count(UV low, UV high);
+extern UV  twin_prime_count_approx(UV n);
+extern UV  nth_twin_prime(UV n);
+extern UV  nth_twin_prime_approx(UV n);
 
 extern int powerof(UV n);
 extern int is_power(UV n, UV a);
